/* $Workfile: A66GroundStart.c $       $Revision: 9 $      $Modtime: 12/14/05 6:12p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    A66GroundStart.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board with the optional ground
    start circuit.
    
    The application provides a simple console interface to manage and
    demonstrate the ground start device.
    
    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxData.h"



/*******************************************************************************
  Macros, Types
*******************************************************************************/

#define DEMO_TASK_REF           1L   // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "A66GroundStart.log"   // Name of the LOG file
#define DEMO_DEFAULT_BOARD_NO   0


typedef struct DemoEnvT {
  char        *pszInputStr;
  OTX_BOOL     bExitOnError;
  char        *logFileName;
  int          boardNo;  
  OTX_INT32    nBoardType;
} DemoEnvT;



/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void *DemoEventFunction( void *lpParameter  /* thread data */ ); 
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
void DemoExitQuestion(void);

void DemoMain(); // defined in Demo.c

                            
/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

OTX_UINT32 g_enableFXO = 0x3F;

static OsaMutexS g_demoLogfCritSec;

struct {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hTss0;
} g_hPhysDevs;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); 
  DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:
  case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:
  case OTX_DEVICE_ARNI_6X3_PCI_PLUS:
    DemoLogf("Arni-6-FXO-PCI-Plus, Arni-6-FXS-PCI-Plus, Arni-6x3-PCI-Plus\n");
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", 
             g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, 
                                    (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 
                                    g_demoEnv.boardNo, 
                                    g_demoEnv.boardNo, 
                                    g_hEventQueue, 
                                    &g_hPhysDevs.hBoardController0); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }

  // Get a handle to the Time-Space Switch (TSS) on the board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                    OTX_DEVICE_TSS, 
                                    0, 
                                    0, 
                                    g_hEventQueue, 
                                    &g_hPhysDevs.hTss0); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(g_hPhysDevs.hTss0, result);
}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0, OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(g_hPhysDevs.hBoardController0, result);

}


void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if they want to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, 
                                     OTX_ATTR_DEVICE_TYPE, 
                                     &eAttribType, 
                                     &cbAttribValue, 
                                     &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, 
                                     OTX_ATTR_DEVICE_NO, 
                                     &eAttribType, 
                                     &cbAttribValue, 
                                     &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", 
               OTX_RESULT_CODE(result), 
               OTX_RESULT_SOURCE_ID(result), 
               OTX_RESULT_LINE_NUM(result), 
               hDevice, 
               (int) eDeviceType, 
               (int)nDeviceNumber, 
               msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("\nDo you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }

}

void DemoPrintTimeStamp( OTX_DATETIME nTimeStamp )
/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds 
  // (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}

void DemoPrintEvent( OtxEventDataS *pEventData )
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_UINT32 readValue;

  switch (pEventData->m_nCode) {

    case OTX_BRD_EC_GROUND_START:
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      DemoLogf("Ground Start Event Received 0x%X\n", pEventData->m_nCode);

      // Read FPGA interrupt register to verify GS interrupt is still active
      // and not a false signal.
      result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                               0x18,   // FPGA Interrupt Pending
                               &readValue);
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("FPGA interrupt pending %#X\n", readValue);
      if(readValue & 0x08000) {
        DemoLogf("GS interrupt pending\n");
      }
      else {
        DemoLogf("GS interrupt not pending\n");
      }

      // Get status from GS device
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             &readValue); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("GS status was %d (%#X)\n", readValue, readValue);

      if(readValue & 0x01) {
        DemoLogf("Ground detected on TIP, line 0\n");
      }

      if(readValue & 0x02) {
        DemoLogf("Ground detected on TIP, line 1\n");
      }

      if(readValue & 0x04) {
        DemoLogf("Ground detected on TIP, line 2\n");
      }

      if(readValue & 0x08) {
        DemoLogf("Ground detected on TIP, line 3\n");
      }

      if(readValue & 0x10) {
        DemoLogf("Ground detected on TIP, line 4\n");
      }

      if(readValue & 0x20) {
        DemoLogf("Ground detected on TIP, line 5\n");
      }

      break;

    default:
      // Unhandled event
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_nCode);
      break;
  }
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoEventFunction( void *lpParameter  /* thread data */ ) 
{
  OTX_RESULT result;
  OtxEventDataS eventData;

  OsaThreadSleep(300); // Give hardware time to startup and initialize

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); 
      DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {

            default:
            break;
        }

        // Printf the event data.
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}

/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    

      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if        (OsaToUpper(*(argv[i] + 2)) == 'E') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'F') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'G') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else {            
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[E|F|G     : Board Type used for the demo, where\n");
  DemoLogf("                    E  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    F  = Use Arni-6-FXO-PCI-Plus Board\n");
  DemoLogf("                    G  = Use GROUND_START Board\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32  exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("A66GroundStart, Ground Start Demo for "
           "Arni-6x6-PCI and Arni-6x6-PCI-Plus\n");

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, 
                                         OTX_ATTR_LIB_TYPE_STATIC, 
                                         &bIsStaticLinkedLibrary); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_FILE_NAME, 
                                       szLibFileName, 
                                       64); 
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", 
             szVersionInfo, 
               szLibFileName, 
               bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library."
             "\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision."
             "\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }


  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-6x6-PCI board
  DemoOpenPhysicalDevices();

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, 
                                    OTX_ATTR_BRD_CPLD_BUILD, 
                                    &eAttribType, 
                                    &cbAttribValue, 
                                    &bReadOnly); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, 
                                       OTX_ATTR_BRD_CPLD_BUILD, 
                                       pszCpldBuildNumbers, 
                                       cbAttribValue); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);

  }

  // Enable Physical devices. Device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoEventFunction);
  OsaThreadSleep(300); // Only so the following message will show on the screen
  DemoLogf("\nPress 'h' for help\n");


  while (!exitFlag) 
  {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Delete allocated resources
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{

  OTX_RESULT result;
  OTX_UINT32 readValue;
  char strBuf[256];

  DemoLogf("User command > %c (0x%x)\n", commandChar, commandChar);

  switch (commandChar) 
  {
  case '?':
  case 'h':
  case 'H':
    DemoLogf("\nCmd  Synopsis for Ground Start\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit this menu\n");
    DemoLogf(" a   Reset GS Circuit\n");
    DemoLogf(" A   Remove reset from GS Circuit\n");
    DemoLogf(" b   Enable GS Circuit\n");
    DemoLogf(" B   Disable GS Circuit\n");
    DemoLogf(" c   Configure GS and enable all lines\n");
    DemoLogf(" d   Display GS Configuration\n");
    DemoLogf(" e   Enable RING Ground on a single line\n");
    DemoLogf(" E   Disable RING Ground on a single line\n");
    DemoLogf(" F   Set FXO line for Off hook operation\n");
    DemoLogf(" r   Read GS status Register\n");
    DemoLogf(" w   Write GS status Register\n");
    DemoLogf(" x   Read FPGA Register\n");
    DemoLogf(" y   Write FPGA Register\n");
    DemoLogf("\n");
    break;

  case 'a':
    DemoLogf("GS is held in reset\n");
    // Reset GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_RESET,  
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    break;

  case 'A':
    DemoLogf("GS reset signal removed (Active)\n");
    // Remove reset from GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_RESET,  
                                         OTX_FALSE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    break;

  case 'b':
    DemoLogf("GS interrupt mask enabled\n");
    // Enable GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_ENABLED,  
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    // Set global FXO enables back to all enabled. 
    g_enableFXO = 0x3F;
    // Enable GS detection on all of the channels
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_GROUND_START_STATUS,  
                                           g_enableFXO); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    break;

  case 'B':
    DemoLogf("GS interrupt mask disabled\n");
    // Disable GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_ENABLED,  
                                         OTX_FALSE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    break;

  case 'c': // Configure Arni board for GS detection

    // Cycle reset on GS circuit
    DemoLogf("Cycle GS reset\n");

    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_RESET,  
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    // Remove reset from GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_RESET,  
                                         OTX_FALSE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);    

    // Enable GS on each of the channels
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_GROUND_START_STATUS,  
                                           g_enableFXO); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    
    DemoLogf("GS interrupt mask enabled\n");
    // Enable GS circuit
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                                         OTX_ATTR_BRD_GROUND_START_ENABLED,  
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("Configuration finished\n");
    // Fall thourgh to show user what has been enabled
  case 'd':
    // Display the GS configuration
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                             0x10,   // FPGA Interrupt Mask
                             &readValue);
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("\nFPGA interrupt mask %#X\n", readValue);
    if(readValue & 0x08000) {
      DemoLogf("GS interrupt mask enabled\n");
    }
    else {
      DemoLogf("GS interrupt mask disabled\n");
    }

    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                             0x18,   // FPGA Interrupt Pending
                             &readValue);
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("FPGA interrupt pending %#X\n", readValue);
    if(readValue & 0x08000) {
      DemoLogf("GS interrupt pending\n");
    }
    else {
      DemoLogf("GS interrupt not pending\n");
    }

    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                             0x30,   // FPGA Reset
                             &readValue);
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("FPGA Reset %#X\n", readValue);
    if(readValue & 0x080) {
      DemoLogf("GS is being reset\n");
    }
    else {
      DemoLogf("GS is not in reset\n");
    }

    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                             0x0,   // FPGA Status
                             &readValue);
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("FPGA Status %#X\n", readValue);
    if(readValue & 0x80000000) {
      DemoLogf("Non-critical interrupts are enabled\n");
    }
    else {
      DemoLogf("Non-critical interrupts are disabled\n");
    }

    DemoLogf("\n");

    break;

  case 'e': // Enable Ground on RING
    DemoLogf("Enable Ground on RING\n");

    DemoLogf("Enter line number (0-5) to enable Ground on RING: ");
    OsaConsoleGetStr(strBuf);

    readValue = strtoul(strBuf, NULL, 0);

    if(readValue < 6) {
      OTX_UINT32 data = 0x01;

      // Enable bits for Ground RING are 6-11 of Ground Start Status Register
      data = data << (readValue + 6);

      // The Ground RING bits in the status register are R/W. Read the current
      // setup from the status register.
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             &readValue); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      // Enable the RING ground and make sure the GS line enable is also
      // enabled. The GS status register bits 0-5 are write only for the
      // GS line enable and read only for the GS detection.It is up to 
      // the application to keep track of which GS line enables should 
      // be enabled.
      data = (data | readValue) | g_enableFXO;

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             data); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    }
    else {
      DemoLogf("Not a valid Line number. Should be(0-5). Not %d\n", readValue);
    }

    break;

  case 'E': // Disable Ground on RING
    DemoLogf("Disable Ground on RING\n");

    DemoLogf("Enter line number (0-5) to disable Ground on RING: ");
    OsaConsoleGetStr(strBuf);

    readValue = strtoul(strBuf, NULL, 0);

    if(readValue < 6) {
      OTX_UINT32 data = 0x01;

      // disable bits for Ground RING are 6-11 of Ground Start Status Register
      data = data << (readValue + 6);
      // The Ground RING bits in the status register are R/W. Read the current
      // setup from the status register.
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             &readValue); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);   

      // Disable the RING ground and make sure the GS line enable is
      // enabled. The GS status register bits 0-5 are write only for 
      // the GS line enable and read only for the GS detection. It is
      // up to the application to keep track of which GS line enables
      // should be enabled.
      data = (~data & readValue) | g_enableFXO;

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             data); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    }
    else {
      DemoLogf("Not a valid Line number. Should be(0-5). Not %d\n", readValue);
    }

    break;

  case 'f': // Setup FXO line for Off hook operation

    DemoLogf("Set FXO line for OFF HOOK state\n");

    DemoLogf("Enter line number (0-5): ");
    OsaConsoleGetStr(strBuf);

    readValue = strtoul(strBuf, NULL, 0);

    if(readValue < 6) {
      OTX_UINT32 data = 0x01;

      // GS detection enable bits are (5-0)
      data = data << readValue;
      // clear only a single line in the global FXO enable mask
      g_enableFXO = g_enableFXO & ~data;

      // The Ground RING bits in the status register are R/W. Read the current
      // setup from the status register.
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             &readValue); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);   

      // Make sure to clear the detection bits. They are bits (5-0) when
      // reading the GS status. These bits are used as the GS detection 
      // enable.
      readValue = readValue & ~(0x3FL);

      // Disable GS on the line that is Off hook. Disable/enable bits for 
      // Ground RING are 6-11 of Ground Start Status Register
      data = data << 6;

      // Disable the RING ground and make sure the GS line enable is
      // enabled. The GS status register bits 0-5 are write only for 
      // the GS detection enable and read only for the GS detection. 
      // It is up to the application to keep track of which GS line 
      // enables should be enabled.
      data = (~data & readValue) | g_enableFXO;

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                             OTX_ATTR_BRD_GROUND_START_STATUS,  
                                             data); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }
    else {
      DemoLogf("Not a valid Line number. Should be(0-5). Not %d\n", readValue);
    }

    DemoLogf("Done\n");

    break;

  case 'r':
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_GROUND_START_STATUS,  
                                           &readValue); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("\nGS status was %d (%#X)\n", readValue, readValue);

    break;
  case 'w':
    DemoLogf("Enter data to be written to GS hardware: ");
    OsaConsoleGetStr(strBuf);

    readValue = strtoul(strBuf, NULL, 0);

    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_GROUND_START_STATUS,  
                                           readValue); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    DemoLogf("Done\n");

    break;

  case 'x':
    {
      OTX_UINT32 regReadAddL = 0;
      OTX_UINT32 regValueL = 0;

      DemoLogf("Enter FPGA register (address) to be read : ");
      OsaConsoleGetStr(strBuf);

      regReadAddL = strtoul(strBuf, NULL, 0);

      result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0,
                               regReadAddL,
                               &regValueL);
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("\nRegister value %d (%#X)\n", regValueL, regValueL);
    }
  break;

  case 'y':
    {
      OTX_UINT32 regReadAddL = 0;
      OTX_UINT32 regValueL = 0;

      DemoLogf("Enter FPGA register (address) to be written : ");
      OsaConsoleGetStr(strBuf);

      regReadAddL = strtoul(strBuf, NULL, 0);

      printf("Enter value to be written ");
      OsaConsoleGetStr(strBuf);

      regValueL = strtoul(strBuf, NULL, 0);

      result = OtxDrvWriteReg32(g_hPhysDevs.hBoardController0, 
                                regReadAddL, 
                                regValueL); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("Done\n");
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}

